import os.path

import numpy as np
import pytest

import horos_io._utils as _utils
from horos_io import load_cine_sequence, _config


@pytest.mark.parametrize("f", ["cmr.py", "cli.py", "_utils.py"])
def test_glob_ssf(f):
    paths = _utils.globSSF("*", root_dir="horos_io")
    assert f in paths


@pytest.mark.parametrize("i, i_expected", [(9, "09"), pytest.param(-1, ...),
                                           pytest.param(100, ...)])
def test__to_str(i, i_expected):
    if i_expected is not ...:
        assert _utils._to_str(i) == i_expected
    else:
        with pytest.raises(ValueError):
            _utils._to_str(i)


def test__has_dicom(horos_test_seq_path):
    assert _utils._has_dicom(horos_test_seq_path)
    assert not _utils._has_dicom(".")


def test_get_ids(horos_test_root):
    for ID, target_ID in zip(_utils.get_ids(horos_test_root), ["Impression_Cmr0064", "Impression_Cmr0067"]):
        assert ID == target_ID


def test_mask_from_omega_contour(horos_test_seq_path):
    """passing a mix of LAX and SAX sequences and testing, if shape is of loaded contours 1D vs 2D is respected"""
    # for p in horos_test_seq_path:
    p = horos_test_seq_path
    cine = load_cine_sequence(p)

    def random_contours():
        result = np.zeros_like(cine, dtype=object)
        Q = result.flatten()
        Q[0] = [(np.random.random(), np.random.random()) for _ in range(np.random.randint(5, 20))]
        result = Q.reshape(result.shape)
        return result

    omega = {cname: random_contours() for cname in _config.omega_4ch_names}

    mask = _utils.mask_from_omega_contour(cine, omega, (0, 0))

    assert mask.shape == cine.flatten()[0].pixel_array.shape




import glob
import os
import os.path
import pathlib
import sys
import warnings
from typing import List, Dict, Union, Optional, Tuple

import cv2
import numpy as np
import pandas as pd
import pydicom
from decorator import decorator
from pydicom.errors import InvalidDicomError
from scipy import interpolate

from horos_io import _config

Path = Union[str, os.PathLike]


def __always_true(*args, **kwargs) -> bool: return True


@decorator
def handle_1D(f, *args, **kwargs):
    def _handle_1D(item):
        if isinstance(item, np.ndarray):
            if len(item.shape) < 2:
                return item[:, np.newaxis]
            else:
                return item
        elif isinstance(item, dict):
            return {k: _handle_1D(v) for k, v in item.items()}
        elif isinstance(item, tuple):
            return tuple([_handle_1D(v) for v in item])
        else:
            return item

    return f(*_handle_1D(args), **_handle_1D(kwargs))


def _to_str(i):
    if not 99 >= i >= 0 or not round(i) == i:
        raise ValueError(f"i should be 2 digits, but instead it is {i}")
    return str(i) if i >= 10 else f"0{i}"


def get_log(log: str, root: str):
    if log is None:
        log = os.path.join(root, "val_contour_log.csv")
    if os.path.isfile(log):
        df = pd.read_csv(log, index_col=0)
    else:
        df = pd.DataFrame(columns=["ID", "contour_type", "result", "by", "frame", "slice", "remark", "time_stamp"])
    return df


def _has_dicom(pth: Path) -> bool:
    for f in glob.glob(os.path.join(pth, "*")):
        try:
            pydicom.dcmread(f)
            return True
        except InvalidDicomError:
            continue
        except PermissionError:
            continue
    return False


def get_seq_paths(root: Path) -> List[str]:
    """
    checks for sequences, which are assumed to  be 3 levels into the root;

    Args:
        root: path to the Horos Dataset
    Returns:
        list of sequence names
    """
    return [pth for pth in globSSF("*/*/*/", root_dir=root) if _has_dicom(os.path.join(root, pth))]


def get_ids(root: Path):
    """
    gets a list of IDs; not as trivial, as there can be other folders, too;
    checks for sequences, which are assumed to  be 3 levels into the root;
    splits the first directory of it and returns non duplicates of it
    Args:
        root: path to the Horos Dataset
    Returns:
        list of IDs
    """
    seqs = get_seq_paths(root)
    result = [pathlib.Path(seq).parts[0] for seq in seqs]
    return sorted(list(set(result)))


def get_log_path(log: str, root: str):
    if log is None:
        log = os.path.join(root, "val_contour_log.csv")
    return log


def globSSF(path_name, /, root_dir=None, **kwargs) -> List[str]:
    """mimcks behaviour of future globs"""
    if root_dir is None:
        return glob.glob(path_name, **kwargs)

    return [os.path.relpath(pth, root_dir) for pth in
            glob.glob(os.path.normpath(os.path.join(root_dir, path_name)), **kwargs)]


####
def mask_from_omega_contour(cines: np.ndarray, contours: Dict[str, np.ndarray], loc: Union[int, Tuple]) -> np.ndarray:
    # TODO: raises error when passing 1D sequences
    # simply havin the user pass the location as tuple or int should be much more flexible
    mask = np.zeros_like(cines[loc].pixel_array)
    for i, (name, c) in enumerate(contours.items()):
        if name != "aroot":
            tck, _ = interpolate.splprep([*zip(*c[loc])], s=0, per=True)
            xnew, ynew = interpolate.splev(np.linspace(0, 1, 500), tck, der=0)
            smooth_contour = np.array([(round(px), round(py)) for px, py in zip(xnew, ynew)])
            mask = cv2.drawContours(mask, [smooth_contour], 0, i + 1, - 1)
        else:
            aroot_c = interpolate_aroot(c[loc])
            mask = cv2.drawContours(mask, [np.array(aroot_c, dtype=int)], 0, i + 1, -1)
    return mask


def interpolate_aroot(landmarks: List[tuple]) -> List[tuple]:
    """
    an aortic root selection of landmarks consists of exactly 6 points; starting with 2 distal points;
    and then going clockwise;

    interpolates the lateral segments, while using linspace for the horizontal sections;

    Args:
        landmarks:

    Returns:

    """
    if not len(landmarks) == 6:
        raise ValueError(f"aroot needs to have exactly 6 landmarks; instead there are {len(landmarks)}")

    def linspace_interpolation(p1, p2, n=10):
        result = np.linspace(0, 1, n)[:, np.newaxis] * (np.array(p2) - p1) + p1
        result = [(x, y) for x, y in result]
        return result

    def spline_interpolation(seq: List[tuple], n=10) -> List[tuple]:
        """use basis shift and do a natural cubic spline interp. and shift back"""
        new_base = np.column_stack([
            v1 := np.array(seq[-1]) - seq[0],
            [-v1[1], v1[0]]
        ])
        transformed_seq = (np.linalg.inv(new_base) @ np.array(seq).T).T
        sp = interpolate.CubicSpline(*transformed_seq.T, bc_type="natural")
        new_seq = sp(np.linspace(transformed_seq[0, 0], transformed_seq[-1, 0], n)) # take x coords of seq endpoints
        result = [(x, y) for x, y in (new_base @ new_seq.T).T]
        return result

    dist = linspace_interpolation(landmarks[0], landmarks[1])
    inner = spline_interpolation(landmarks[1:4])
    prox = linspace_interpolation(landmarks[3], landmarks[4])
    outer = spline_interpolation(landmarks[-2:] + [landmarks[0]])
    return dist + inner + prox + outer


def masks2spline(masks: np.ndarray):
    """converts an integer mask into a list of contours"""
    return [masks2spline(masks == i) for i in np.unique(masks)]


def mask2spline(mask, ):
    """
    Convert a binary mask to a spline.

    Args:
        mask (ndarray): Binary mask.
        epsilon (float): The epsilon value for RDP algorithm.

    Returns:
        x (ndarray): x coordinates of the spline.
        y (ndarray): y coordinates of the spline.
    """
    # Prepare the mask
    warnings.warn(DeprecationWarning("dont use horos_io.mask2spline for now..."))
    if not mask.any():
        return np.nan
    _, thrs = cv2.threshold(mask, 0, 255, cv2.THRESH_OTSU)

    # Find the contour
    contours, _ = cv2.findContours(thrs, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE)
    sorted_contours = sorted(contours, key=cv2.contourArea, reverse=True)
    largest_contour = sorted_contours[0]
    largest_contour = [(p[0], p[1]) for p in largest_contour]

    return largest_contour


def get_tag(tag: Tuple[str, str], ID: str, root: str) -> str:
    dicom_paths = globSSF(f"{os.path.join(root, ID, '*/*/*.dcm')}")
    return pydicom.dcmread(dicom_paths[0])[tag].value


def get_study_date(ID: str, root: str) -> str:
    return get_tag(_config.study_date_tag, ID, root)



import tkinter as tk
from tkinter import ttk

from matplotlib import pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


def shitty_confirmation_ui(ax: plt.Axes) -> tk.Tk:
    window = tk.Tk()
    window.geometry("+%d+%d" % (50, 50))
    image_window = FigureCanvasTkAgg(ax.figure, window)
    image_window.get_tk_widget().grid(column=0, row=0, columnspan=2, rowspan=2)
    confirm = ttk.Button(window, text="CONFIRM",
                         command=lambda: (setattr(window, "truth", True),
                                          setattr(window, "remark", remark.get("1.0", "end")),
                                          window.quit(),
                                          window.destroy()))
    confirm.grid(column=0, row=2)
    deny = ttk.Button(window, text="DENY",
                      command=lambda: (setattr(window, "truth", False),
                                       setattr(window, "remark", remark.get("1.0", "end")),
                                       window.quit(),
                                       window.destroy()))
    deny.grid(column=1, row=2)
    remark_label = tk.Label(window, text="Remarks: ")
    remark_label.grid(column=2, row=0)
    remark = tk.Text(window)
    remark.grid(column=2, row=1, rowspan=2)
    return window


def shitty_manual_confirm(sth: plt.Axes) -> bool:
    """
    plots an image to a window and waits for you to confirm or deny... yep; creates a new fucking window every time
    """
    window = shitty_confirmation_ui(sth)
    window.mainloop()
    return window.truth, window.remark



import os.path

import pandas as pd
import pytest


@pytest.fixture()
# if testing horos; with our data, change this path to where your data is located
def horos_test_root():
    return os.path.normpath("./tests/horos_dummy")

SEQ = ['Impression_Cmr0064/Anonymous_Study - 0/CINE_segmented_LAX_2Ch_7/',
                        'Impression_Cmr0064/Anonymous_Study - 0/CINE_segmented_LAX_3Ch_5/',
                        'Impression_Cmr0064/Anonymous_Study - 0/CINE_segmented_LAX_4Ch_6/',
                        'Impression_Cmr0064/Anonymous_Study - 0/CINE_segmented_SAX_28/',
                        'Impression_Cmr0067/Anonymous_Study - 0/CINE_segmented_LAX_2Ch_7/',
                        'Impression_Cmr0067/Anonymous_Study - 0/CINE_segmented_LAX_3Ch_9/',
                        'Impression_Cmr0067/Anonymous_Study - 0/CINE_segmented_LAX_4Ch_10/',
                        'Impression_Cmr0067/Anonymous_Study - 0/CINE_segmented_SAX_53/']
N_SLICES = [1, 1, 1, 13, 1, 1, 1, 13]

@pytest.fixture(params=SEQ)
def horos_test_seq_path(horos_test_root, request):
    return os.path.normpath(os.path.join(horos_test_root, request.param))


@pytest.fixture(params=N_SLICES)
def horos_test_seq_n_slices(request):
    return request.param


@pytest.fixture()
def horos_basal_first_file(horos_test_root):
    return os.path.join(horos_test_root, "basal_info.csv")


@pytest.fixture()
def horos_contour_info_path(horos_test_root):
    return os.path.join(horos_test_root, "target_contour_info.csv")


@pytest.fixture()
def horos_image_info_path(horos_test_root):
    return os.path.join(horos_test_root, "target_image_info.csv")


@pytest.fixture()
def log_dummy():
    return pd.read_csv("./tests/visual_confirmation.csv", index_col=0)

# def pytest_generate_tests(metafunc: pytest.Metafunc):
#
#     if "horos_test_seq_path" in metafunc.fixturenames and "horos_test_seq_n_slices" in metafunc.fixturenames:
#         metafunc.parametrize("horos_test_seq_path, horos_test_seq_n_slices", zip(SEQ, N_SLICES))


def pytest_addoption(parser):
    parser.addoption("--pth")



"""
Console script for horos io;
will incorporate functionality for running visual sanity checks on contours overlayed on top of the data;

generating .csv files for data
"""
import os
import sys
import typing
from typing import Optional, NoReturn

import click
import pandas as pd
from tqdm import tqdm

from horos_io import load_horos_contour
from horos_io._utils import globSSF, mask_from_omega_contour, get_log, get_log_path
from horos_io.cmr import Path, get_image_info, get_contour_info, get_combined_info
from horos_io.ui import shitty_manual_confirm
from horos_io.validation import visually_confirm_omega_iter, write_log, last_validation_was_successful, \
    contour_is_valid, changed_since_last_validation, last_validation
import matplotlib
if sys.platform == "darwin":
    matplotlib.use("macosx")

from matplotlib import pyplot as plt

@click.group()
def main():
    """Console script for horos_io."""
    return 0


@click.command("image_info",
               help="saves information about the horos dataset as csv, such that it can be later used for efficient"
                    "loading of the data")
@click.option("--root", default=".", help="directory to the Horos Dataset")
@click.option("--silent", default=False, is_flag=True, help="if set, will not print out the result")
@click.option("--out", default=None,
              help="how the image_info,csv should be saved; if not specified, will save an image_info.csv in root")
def make_image_info_csv(root: Path,
                        out: Optional[Path],
                        silent: bool) -> NoReturn:
    """
    one-time operation to update the information of the dataset;
    columns:
    ("seq_path",  <relative path starting from the root of the data>
     "location",  joins the rootdir with seq path
     "ID",        <Impression_Cmr{ID}>
     "slice_type", one of config.slice_types
     "n_frames",
     "n_slices",

    seq_path is relative to the root
    location is joint path of root and seq_path

    creates a csv file containing location of the sequence.
    sequence can be loaded with load_cine_sequence
    Args:
        silent: if silent, will not print out the result to console
        root: root of the data; contents should be directories of Impression Studies
        out: optional parameter to specify, what and where to save; if None, will save image_info.csv inside root
    Returns:
    """
    click.echo(f"retrieving image info from {root}")
    result = get_image_info(root)
    if not silent:
        click.echo("Findings:\n\n")
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):  # more options can be specified also
            click.echo(result)
    if out is None:
        out = os.path.normpath(os.path.join(root, "image_info.csv"))
    click.echo(f"saving result to {out}")
    result.to_csv(out)


@click.command("contour_info",
               help="saves information about the horos dataset as csv, such that it can be later used for efficient"
                    "loading of the data")
@click.option("--root", default=".", help="directory to the Horos Dataset")
@click.option("--silent", default=False, is_flag=True, help="if set, will not print out the result")
@click.option("--out", default=None,
              help="how the image_info,csv should be saved; if not specified, will save an image_info.csv in root")
def make_contour_info_csv(root: Path, out: Optional[Path], silent: bool) -> NoReturn:
    """
    one-time operation to update the information of the dataset;

    ("ID": <Impression_Cmr{ID}>,
     "contour_path": relative path to the contour file, starting from root of the data,
     "contour_type": one of config.contour_types,
     "location": joins the root dir with the sequence path
     "slice_type": one of config.slice_types,
    )

    creates a csv file containing location of all contour files; or whether they exist at all.
    Args:
        root: root of the data; contents should be directories of Impression Studies
    Returns:
    """
    root = os.path.normpath(root)
    click.echo(f"creating contour info for {root}")
    result = get_contour_info(root)
    if not silent:
        click.echo("Findings:\n\n")
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):  # more options can be specified also
            click.echo(result)

    if out is None:
        out = os.path.normpath(os.path.join(root, "contour_info.csv"))
    click.echo(f"saving contour info to: {out}")
    result.to_csv(out)


@click.command()
@click.option("--by", required=True, help="the name that will be recorded in the log_")
@click.option("--root", default=".", help="source root of the Horos dataset")
@click.option("--only_unconfirmed", is_flag=True, default=False,
              help="will only prompt you cases, that have not been confirmed by the user yet (registered in log_)")
@click.option("--log", default=None,
              help="if None, will save or update a val_contour_log.csv in root, else checks your specified log_ instead")
@click.option("--dry_run", is_flag=True, default=False,
              help="if this flag is set, run through the procedure without displaying or logging anything")
def validate(by: str, root: Path, only_unconfirmed: bool, log: Path, dry_run: bool):
    """
    visually confirm literally every existing omega contour... (might need to change with params to return
    an adequate iterator)

    will also save a dataframe / update a dataframe in <root>/val_contour_log.csv,
    that logs review result with timestamp
    """
    click.echo(f"starting validation process in root: {root}")

    for ID, omega, f, s, cines, contours, mt_contour in tqdm(
        visually_confirm_omega_iter(combined_info=get_combined_info(root))):
        if only_unconfirmed and last_validation_was_successful(get_log(log, root), ID=ID, frame=f, slice=s,
                                                               contour_type=omega) and not changed_since_last_validation(
            mt_contour, last_validation(get_log(log, root), ID=ID, frame=f, slice=s,
                                        contour_type=omega)):  # CAVE: hard coded
            click.echo(f"skipping {ID}, ({f}, {s}), {omega}")
            continue
        loc = (f, s)
        fig, ax = plt.subplots()
        ax.imshow(cines[loc].pixel_array)
        ax.imshow(mask_from_omega_contour(cines, contours, loc), cmap="jet", alpha=0.5)
        ax.figure.suptitle(f"{ID}: {omega} f: {f}, s: {s}, ")

        ##################### ideall y this is not part of the cli logic #######################################
        if omega == "omega_3ch":
            ax.axes.scatter(*zip(*contours["aroot"][loc]),
                            c=["black", "red", "green", "blue", "cyan", "purple"])
        if not dry_run:
            ok, remark = shitty_manual_confirm(ax)
            write_log(log, **locals())
        else:
            plt.close()

    click.echo("batch finished validation!")


@click.command("rename", help="renames all contours with .xml and .rois_series attached to it")
@click.argument("old", nargs=1)
@click.argument("target", nargs=1)
@click.option("--root", help="Data root")
def rename_contour_files(root: str, old: str, target: str) -> typing.NoReturn:
    """
    renames .xml and .roi_series files by dictionary, that lie in the impression folders:
    root |--- Impression_CmrXXXX
                    |---- lv_epi.xml
                    |---- lv_epi.roi_series
    Args:
        target:
        old:
        root:
    Returns:
    """
    files = globSSF(f"*/{old}.xml", root_dir=root) + globSSF(f"*/{old}.rois_series", root_dir=root)
    for f in files:
        path, f_name = os.path.split(f)
        f_name, f_ending = f_name.split(".")
        src = os.path.normpath(os.path.join(root, path, f"{old}.{f_ending}"))
        dst = os.path.normpath(os.path.join(root, path, f"{target}.{f_ending}"))
        if os.path.isfile(dst):
            raise FileExistsError(f"{dst} for {src} already exists")
        os.rename(src, dst)
        click.echo(f"renamed {src} to {dst}")
    click.echo("finished")


@click.group("check", help="given a root and a log file, checks if the user (--by) has seen all the samples")
def check():
    pass


@click.option("--by", required=True, help="the name to check for in the logs")
@click.option("--root", default=".", help="source root to the Horos dataset")
@click.option("--log", default=None,
              help="if None, will check for val_contour_log.csv in root, else checks your specified log instead")
@click.command("failed", help="check to log for which contours did not and in validation by the user")
def check_failed(by: str, root: str, log: str):
    click.echo(
        f"checking for failed validations by {by} for dataset in {root}, using logfile {get_log_path(log, root)}")
    log_df = get_log(log, root)
    failed = log_df.apply(lambda row: not last_validation_was_successful(log_df,
                                                                         by=row["by"],
                                                                         ID=row["ID"],
                                                                         contour_type=row["contour_type"],
                                                                         frame=row["frame"],
                                                                         slice=row["slice"]), axis=1)
    result = log_df[(log_df["by"] == by) & failed]
    result = result[["by", "ID", "contour_type", "frame", "slice", "remark"]].drop_duplicates()
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        click.echo(result)
        click.echo("----------------")
        click.echo(f"These samples did not end with confirmation by user {by}")


@click.command("unobserved",
               help="looks at the data and the logs to determine, which contours by which ID still needs to,"
                    "be evalutated by a given user;")
@click.option("--root", default=".", help="source root to the Horos dataset")
@click.option("--log", default=None,
              help="if None, will check for val_contour_log.csv in root, else checks your specified log instead")
@click.option("--by", required=True, help="the name to check for in the logs")
def check_unobserved(by: str, root: str, log: str):
    click.echo(
        f"checking, which contours have not been seen by user {by} yet in datasat at {root}, using log {get_log_path(log, root)}")
    log_df = get_log(log, root)
    combined_info_ = get_combined_info(root)
    # combined_info_["ID", "contour_type"]
    click.echo("Under construction")
    return

    # for g_n, group in combined_info_.groupby(["ID", "slice_type", "contour_type", "location_contour", "location_images"]):
    #     click.echo(group)
    #
    # # TODO: open the contour to look for which frame / slice needs to be seen
    # # TODO: filter by omega


@click.command("existing", help="look at exsiting contours in root, given a number of contour types; will check"
                                "for existing <contour_type>.xml files that are in child directories of root")
@click.argument("contour_types", nargs=-1)
@click.option("--n", default=None, type=int,
              help="If passed, will check if a contour file has at least that many contours")
@click.option("--root", default=".", help="source root to the Horos dataset")
def check_existing_contours(contour_types: typing.List[str], root: str, n: Optional[int], ):
    """currently terrible and unhelpful vaidation process... honestly using pytest was pretty king"""
    click.echo(f"checking contour types {contour_types} in root: {root}...")
    combined_info = get_combined_info(root)

    click.echo("the following contours need to be rechecked: ")
    # TODO: improve logging...
    # really, pytest for this kinda stuff wasnt half bad
    for contour_type in contour_types:
        eligible = combined_info[
            (combined_info["contour_type"] == contour_type) & (combined_info["location_contour"].notna())]
        eligible.loc[:, "valid"] = eligible.apply(
            lambda row: contour_is_valid(load_horos_contour(row["location_contour"], row["location_images"]), n),
            axis=1)
        click.echo(eligible[eligible["valid"] != True][["ID", "contour_path", "contour_type"]])

    click.echo("finished checking")


main.add_command(make_image_info_csv)
main.add_command(make_contour_info_csv)
main.add_command(rename_contour_files)
main.add_command(validate)
main.add_command(check)
check.add_command(check_failed)
check.add_command(check_unobserved)
check.add_command(check_existing_contours)

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import os
import time
from datetime import datetime
from functools import reduce
from typing import Optional, Union, Dict

import numpy as np
import pandas as pd

import horos_io._legacy
import horos_io.core
from horos_io.cmr import Path
from horos_io._config import time_format


def last_validation_was_successful(log: pd.DataFrame, conf_field="result", **criteria):
    """
    Args:
        conf_field: name of the field in log_, from which to draw the confirmation
        **criteria: k: column name in log_; v: specific entry to filter by; either a specific value or a callable that is the applied to the respective series
        log:

    Returns:
        bool: True, if the last entry of this combination in log_ was successful
    """
    eligible = reduce(lambda x, y: x & y,
                      [log[k].apply(v) if callable(v) else log[k] == v for k, v in criteria.items()])
    in_question = log[eligible]
    if in_question.shape[0] == 0:
        return False
    else:
        # take the latest
        latest_idx = in_question["time_stamp"].apply(lambda st: datetime.strptime(st, time_format)).idxmax()
        return in_question.loc[latest_idx, conf_field]


def last_validation(log: pd.DataFrame, **criteria) -> pd.Series:
    """
    Args:
        **criteria: k: column name in log_; v: specific entry to filter by; either a specific value or a callable that is the applied to the respective series
        log:

    Returns:
        bool: True, if the last entry of this combination in log_ was successful
    """
    eligible = reduce(lambda x, y: x & y,
                      [log[k].apply(v) if callable(v) else log[k] == v for k, v in criteria.items()])
    in_question = log[eligible]
    if in_question.shape[0] == 0:
        return None
    else:
        # take the latest
        latest_idx = in_question["time_stamp"].apply(lambda st: datetime.strptime(st, time_format)).idxmax()
        return in_question.loc[latest_idx,]


def visually_confirm_omega_iter(combined_info: pd.DataFrame):
    """
    CAVE: should only be used for parametrization for the following test
    TODO: as more contour types roll out; we should dynamically parametrize this from passed arguments
    TODO: turn this into a fixture
    Args:
        combined_info:
    Returns:
    """
    combined_info = combined_info[combined_info["contour_type"].apply(lambda cont: "omega" in cont)]
    for i, row in combined_info.iterrows():
        cines = horos_io.load_cine_sequence(row["location_images"])
        contours = horos_io.core.load_horos_contour(row["location_contour"], row["location_images"])
        mt_contour = time.localtime(os.stat(row["location_contour"]).st_mtime)

        if len(cines.shape) == 1:
            cines = cines[:, np.newaxis]
            contours = {cn: contour[:, np.newaxis] for cn, contour in contours.items()}
        for f, s in zip(*np.where(list(contours.values())[0] != 0)):
            yield row["ID"], row["contour_type"], f, s, cines, contours, mt_contour


def changed_since_last_validation(mt_contour, log_row: pd.Series) -> bool:
    """TODO: check, if this method even works correctly"""
    if log_row is None:
        return True
    last_val = time.strptime(log_row["time_stamp"], time_format)
    return mt_contour > last_val


def write_log(log_: Path, **kwargs):
    if log_ is None:
        log_ = os.path.join(kwargs["root"], "val_contour_log.csv")
    if os.path.isfile(log_):
        df = pd.read_csv(log_, index_col=0)
    else:
        if not os.path.isdir(os.path.split(log_)[0]):
            os.makedirs(os.path.split(log_)[0])
        df = pd.DataFrame(columns=["ID", "contour_type", "result", "by", "frame", "slice", "remark", "time_stamp"])
    df = df.append(pd.Series({
        "ID": kwargs["ID"],
        "contour_type": kwargs["omega"],
        "result": kwargs["ok"],
        "by": kwargs["by"],
        "frame": kwargs["f"],
        "slice": kwargs["s"],
        "remark": kwargs["remark"],
        "time_stamp": datetime.now().strftime(time_format)
    }), ignore_index=True)
    df.to_csv(log_)


def all_more_than_3_points(contour: np.ndarray) -> bool:
    """any contour should have more than 3 points for meaningful spline interpolation; (even though we only need 3 theoretically)"""
    return all([len(C) > 3 for C in contour.flatten()[np.where(contour.flatten() != 0)]])


def contour_is_valid(contour: Union[Dict[str, np.ndarray], np.ndarray], n: Optional[int] = None) -> bool:
    """this method kinda sucks bc it doesnt tell us which and where"""
    if isinstance(contour, dict):
        return reduce(lambda x, y: x and y, [contour_is_valid(value, n) for value in contour.values()])
    else:
        at_least_n = (contour != 0).sum() >= n if n is not None else True
        return at_least_n and all_more_than_3_points(contour)



import os
import re
from functools import partial
from typing import Callable, Union, Dict, Tuple, List
from xml.etree import ElementTree as ET

import numpy as np
import pandas as pd

from horos_io import _config
from horos_io._utils import __always_true, globSSF, get_ids

Path = Union[str, os.PathLike]


def _filter_by_contour_name(contour_name: str, elem: ET.Element) -> bool:
    """just keep calm... dont worry about indexing"""
    return elem[17].text == contour_name


def _get_name_from_template(path_to_sequence: Path, template: str) -> str:
    result = re.findall(template,
                        "_".join(globSSF("*", root_dir=path_to_sequence)))
    if len(result) <= 1:
        return result[0]
    else:
        raise ValueError(f"More than 1 match fore template {template}:\n\n result: {result}")


def _load_horos_contour(contour_path: Path, n_frames: int, n_slices: int,
                        filter_: Callable[[ET.Element], bool] = __always_true) -> np.ndarray:
    """
    to really understand this file format; getting elements should really be smarter than indexing with integers
    anyway;
    simply loads the contour from the contour path;
    Args:
        contour_path: path to contour .xml file
        n_frames: number of frames of the corresponding sequence
        n_slices: number of slices of the corresponding sequence; if 1-Dim; e.g. LAX, then 1 should be passed
        filter_: takes an element at per-Image level (can have multiple ROIs; and decides on some criterion if it should be taken);
                    CAVE: will overwrite contours if you dont filter for exactly one contour per image
    Returns: ndarray in the shape of original dicoms, where each entry is either 0, if no contour, or a list of pixels (tuples)

    """
    tree = ET.parse(contour_path)
    root = tree.getroot()

    result = np.zeros((n_frames, n_slices), dtype=object)

    for elem in root[0][1]:
        # image level
        image_index = int(elem[1].text)
        f = image_index // n_slices
        s = image_index % n_slices  # if longaxis, s will result in 0

        for roi in elem[5]:
            # multiple contours can exist for every image
            if filter_(roi):
                result[f, s] = [eval(point.text) for point in roi[23]]
    return result if result.shape[1] > 1 else result.reshape(result.shape[0])


def _load_omega_contour(contour_path: Path, n_frames: int, n_slices: int) -> Dict[str, np.ndarray]:
    """
    Args:
        contour_path:
        n_frames:
        n_slices:

    Returns:
    """
    # any contour should be named <Path>/omega_{slice_type}.xml
    omega = os.path.split(contour_path)[1].split(".")[0]
    result = {contour_name: _load_horos_contour(contour_path, n_frames, n_slices,
                                                filter_=partial(_filter_by_contour_name, contour_name))
              for contour_name in getattr(_config, f"{omega}_names")}
    return result


def get_n_frames_from_seq_path(path_to_sequence: Path) -> int:
    """
    get the number of frames from the sequence;
    CAVE: asssumes, that there are only files of the form
        IM-XXXX-XXXX.dcm or IM-XXXX-XXXX-XXXX.dcm inside the folder;
        any other form of the data might silently return wrong results
    Args:
        path_to_sequence:
    Returns:
    """
    # thankfully, we only need to do one look ahead
    files = globSSF("*.dcm", root_dir=path_to_sequence)
    if len(files) == 0:
        raise ValueError(f"no .dcm files found in {path_to_sequence}; current dir is {os.getcwd()}")
    s = re.compile(r"(?<=IM-\d{4}-)\d{4}")
    frame_idxs = s.findall("_".join(files))
    return max([int(frame_idx) for frame_idx in frame_idxs])


def get_n_slices_from_seq_path(path_to_sequence: Path) -> int:
    """
    get the number of slices from the sequence;
    CAVE: asssumes, that there are only files of the form
        IM-XXXX-XXXX.dcm or IM-XXXX-XXXX-XXXX.dcm inside the folder;
        any other form of the data might silently return wrong results
    Args:
        path_to_sequence:
    Returns:

    """
    files = globSSF("*.dcm", root_dir=path_to_sequence)
    s = re.compile(r"IM-\d{4}-\d{4}.dcm")
    if s.match(files[0]):
        # we have long axis images,
        return 1
    else:
        s = re.compile(r"(?<=IM-\d{4}-\d{4}-)\d{4}(?=.dcm)")
        slice_idxs = s.findall("_".join(files))
        return max([int(slice_idx) for slice_idx in slice_idxs])


def load_horos_contour(path_to_contour: Path, sequence: Union[np.ndarray, tuple, Path]) -> Union[
    np.ndarray, Dict[str, np.ndarray]]:
    """
    Args:
        path_to_contour: path should be a contour .xml file
        sequence: sequence can be:
                            - a sequence loaded by load_cine_sequence
                            - a path to a sequence that could be loaded by load_cine_sequence
                            - a tuple (n_frames, n_slices)
                  this is needed in order to infer, what the image_index from the .xml mean
    Returns: an ndarray of either shape (n_frames,) or (n_frames, n_slices)
             depending if it was lax or short axis;
             the entries are 0, if no contour, an instance of List[Tuple[float, flaot]], if it exists,
             to get a mask of existing contours, simply do:
             result != 0

             if an omega contour; will instead return a dict[contournames, contour]
    """
    if len(sequence) == 2:
        n_frames, n_slices = sequence
    elif isinstance(sequence, np.ndarray):
        n_frames, n_slices, *rest = sequence.shape
    else:
        # has to be a path or sth; if not, this should automatically result in an error
        n_frames = get_n_frames_from_seq_path(sequence)
        n_slices = get_n_slices_from_seq_path(sequence)
    if "omega" not in path_to_contour:
        return _load_horos_contour(path_to_contour, n_frames, n_slices)
    else:
        return _load_omega_contour(path_to_contour, n_frames, n_slices)


def load_horos_contour_unstructured(contour_path: Path) -> List[Tuple[int, str, List[Tuple[float, float]]]]:
    """
    sometimes it might be easier to not rely on prior structural information to construct an ndarray
    this function simply loads all existing contours in a given .xml file

    to be seen, if there should be more returned
    Args:
        contour_path: path to the contour .xml file
    Returns:
        a list of all existing contours in a triple (ImageIndex, ROIName, Contour)
    """
    tree = ET.parse(contour_path)
    root = tree.getroot()

    result = []
    for elem in root[0][1]:
        # image level
        image_index = int(elem[1].text)
        for roi in elem[5]:
            contour_name = roi[17].text
            contour = [eval(point.text) for point in roi[23]]
            result.append((image_index, contour_name, contour))
    return result


def get_contour_info_by_type(root: Path, contour_type: str) -> pd.DataFrame:
    """
    given a Horos root and a contour type; returns a dataframe with information
    on given contour in that dataset

    (
    "ID": folder name, where contours reside,
     "contour_path": relative path to the contour file, starting from root of the data,
     "contour_type": one of config.contour_types,
     "location": joins the root dir with the sequence path
     "slice_type": one of config.slice_types,
    )
    Args:
        root: Horos dataset root
        contour_type: a string; will search for <contour_type>.xml in ID directories
    Returns:
    """
    result = pd.DataFrame()
    result["ID"] = get_ids(root)
    result["contour_type"] = contour_type

    result["contour_path"] = result["ID"].apply(
        lambda id_: os.path.join(id_, f"{contour_type}.xml")
        if os.path.isfile(os.path.join(root, id_, f"{contour_type}.xml"))
        else ""
    )
    result["contour_path"].replace("", pd.NA, inplace=True)

    result["location"] = result["contour_path"].apply(
        lambda path: os.path.normpath(os.path.join(root, path)) if pd.notna(path) else ""
    )
    result["location"].replace("", pd.NA, inplace=True)
    result = result[["ID", "contour_type", "contour_path", "location"]]
    return result


def existing_contours_within(group: pd.DataFrame) -> pd.DataFrame:
    """
    TODO: finish this
    takes a group from get_combined_info, that has been properly grouped, such that there is only one row left;
    loads contours and creates a new df, with additional entry for frame and slice, where contours exist;

    if a contour object with multiple contours, will raise an error if there is any discrepancy
    Args:
        group:

    Returns:
    """
    pass



#!/usr/bin/env python
"""
Tests for `horos_io` package.
TODO: move away from legacy loading of contours n shit in tests

contains tests for general functionality of loading contours, images and shit exported from our HOROS Mac;
for tests regarding actual integrity of a horos dataset, refer to test_horos_dataset.py
"""
# pylint: disable=redefined-outer-name
import argparse
import os
import sys

import numpy as np
import pandas as pd
import pydicom
import pytest
from matplotlib import pyplot as plt

import horos_io._legacy
import horos_io.core
from horos_io import load_sax_sequence
from horos_io._utils import globSSF
from tests import conftest


def test_get_n_frames_from_seq_path(horos_test_root):
    for seq_path in globSSF("*/*/*/", root_dir=horos_test_root):
        seq_root = os.path.join(horos_test_root, seq_path)
        n_frames = horos_io.core.get_n_frames_from_seq_path(seq_root)
        assert n_frames == 25, f"n_frames should be 25, but is {n_frames}; failed at {seq_root}"


@pytest.mark.parametrize("horos_test_seq_path2, horos_test_seq_n_slices2", zip(conftest.SEQ, conftest.N_SLICES))
def test_get_n_slices_from_seq_path(horos_test_seq_path2, horos_test_seq_n_slices2, horos_test_root):
    pth = os.path.normpath(os.path.join(horos_test_root, horos_test_seq_path2))
    assert horos_io.core.get_n_slices_from_seq_path(pth) == horos_test_seq_n_slices2


def test_load_lax_sequence_pipeline(horos_test_seq_path):
    if not (
        "2ch" in horos_test_seq_path.lower() or "3ch" in horos_test_seq_path.lower() or "4ch" in horos_test_seq_path.lower()):
        assert True
        return

    sample = horos_io.load_lax_sequence(horos_test_seq_path)
    assert sample.shape == (25,)


def test_load_sax_sequence_pipeline(horos_test_seq_path):
    if not "sax" in horos_test_seq_path:
        assert True
        return
    else:
        sample = horos_io.load_sax_sequence(horos_test_seq_path)
        assert sample.shape == (horos_io.core.get_n_frames_from_seq_path(horos_test_seq_path),
                                horos_io.core.get_n_slices_from_seq_path(horos_test_seq_path))


def test_load_cine_sequence_pipeline(horos_image_info_path):
    image_info = pd.read_csv(horos_image_info_path, index_col=0)

    for i, row in image_info.iterrows():
        sample = horos_io.load_cine_sequence(row["location"])
        for entry in sample.flatten():
            assert isinstance(entry, pydicom.FileDataset)


def test_load_horos_contour(horos_image_info_path, horos_contour_info_path):
    contour_info = pd.read_csv(horos_contour_info_path, index_col=0)
    image_info = pd.read_csv(horos_image_info_path, index_col=0)
    combined = pd.merge(contour_info[contour_info["location"].notna()],
                        image_info, on=["ID", "slice_type"], suffixes=("_contour", "_images"))

    for i, row in combined.iterrows():
        result: np.ndarray = horos_io.core.load_horos_contour(row["location_contour"], row["location_images"])
        assert (result != 0).any()


@pytest.mark.skipif(not ("visual" in sys.argv),
                    reason="should only be run, if we want to manually sanitiy check the ordering")
@pytest.mark.visual
def test_visually_confirm_ordering(horos_image_info_path, horos_contour_info_path):
    """
    fix the skipif shenanigans
    Args:
        horos_image_info_path:
        horos_contour_info_path:

    Returns:
    """
    contour_info = pd.read_csv(horos_contour_info_path, index_col=0)
    image_info = pd.read_csv(horos_image_info_path, index_col=0)
    image_info = image_info[image_info["slice_type"] == "cine_sa"]
    combined = pd.merge(contour_info[contour_info["location"].notna()],
                        image_info, on=["ID", "slice_type"], suffixes=("_contour", "_images"))

    for i, row in combined.iterrows():
        cines = horos_io.load_cine_sequence(row["location_images"])
        contours = horos_io.core.load_horos_contour(row["location_contour"], row["location_images"])

        mid_minus3 = cines.shape[1] // 2 - 3

        plt.imshow(cines[0, mid_minus3].pixel_array)
        plt.scatter(*zip(*contours[0, mid_minus3]))
        plt.show()
    assert True


def test__load_omega_4ch_contour():
    pth = "tests/omega_4ch.xml"
    result = horos_io.core.load_horos_contour(pth, (25, 1))

    for contour_name, contours in result.items():
        locs = np.array((0, 10))  # ED in frame 0 and ES in frame 10
        target = np.zeros_like(contours, dtype=bool)
        target[locs] = True
        assert all((contours != 0) == target)


@pytest.mark.parametrize("sax_path, basal_first",
                         [('Impression_Cmr0064\\Anonymous_Study - 0\\CINE_segmented_SAX_28\\', True),
                          ('Impression_Cmr0067\\Anonymous_Study - 0\\CINE_segmented_SAX_53\\', False)])
def test_sort_sax_by_y(sax_path, horos_test_root, basal_first):
    """
    test to see if we can load SAX Stack in right order without relying on manual confirmation
    """
    pth = os.path.normpath(os.path.join(horos_test_root, sax_path))
    sax_manual = horos_io._legacy.load_sax_sequence(pth, basal_first)
    sax_by_y = horos_io.sort_SAX_by_y(load_sax_sequence(pth))
    for by_y, manual in zip(sax_by_y.flatten(), sax_manual.flatten()):
        assert (by_y.pixel_array == manual.pixel_array).all()


def sax_method_iter():
    parser = argparse.ArgumentParser()
    parser.add_argument("--pth")
    args, _ = parser.parse_known_args()
    if args.pth is not None:
        image_info = pd.read_csv(os.path.join(args.pth, "image_info.csv"), index_col=0)
        for i, row in image_info[image_info["slice_type"] == "cine_sa"].iterrows():
            full_seq_path = os.path.normpath(os.path.join(args.pth, row["seq_path"]))
            yield full_seq_path, row
    else:
        yield None, None  # dummy values


@pytest.mark.skipif(not "--pth" in sys.argv, reason="this test is only to be run when specifying a path to a dataset")
@pytest.mark.parametrize("full_seq_path, row", sax_method_iter())
def test_sort_sax_by_y_hard(full_seq_path, row):
    """yaaay"""
    sax_manual = horos_io.load_sax_sequence(full_seq_path)
    sax_by_y = horos_io.sort_SAX_by_y(load_sax_sequence(full_seq_path))
    for s_manual, s_by_y in zip(sax_manual.flatten(), sax_by_y.flatten()):
        assert (s_manual.pixel_array == s_by_y.pixel_array).all()
